/********************************************************************/
/* DXFIN: DXF file input.                                           */
/********************************************************************/

#include <dxfin.h>
#include <io.h>

#define REVERSED_BYTE_ORDER 0  /* set to 1 on HiLo-oriented machines */

/***************************** CDxfin:: *****************************/

/**************************** constructor ***************************/

CDxfin::CDxfin()
{
	fp = NULL;
}

/**************************** destructor ****************************/

CDxfin::~CDxfin()
{
	Close();
}

/****************************** methods *****************************/

/********************************************************************/
/* Open: opens given file and sets up input.                        */
/********************************************************************/

BOOL CDxfin::Open(const char *fn)
{
	Close();
	group = DXF_FILE;
	if((fp = fopen(fn, "rb")) == NULL)
		return FALSE;
	if(!InitRead())
	{
		Close();
		return FALSE;
	}
	return TRUE;
}

/********************************************************************/
/* Open: assigns given stream to dxf input.                         */
/********************************************************************/

BOOL CDxfin::Open(FILE *_fp)
{
	Close();
	group = DXF_FILE;
	if(_fp == NULL)
		return FALSE;
	fp = _fp;
	if(!InitRead())
	{
		fp = NULL;
		return FALSE;
	}
	return TRUE;
}

/********************************************************************/
/* Close: closes DXF file (if open).                                */
/********************************************************************/

void CDxfin::Close()
{
	if(fp)
	{
		fclose(fp);
		fp = NULL;
	}
}

/********************************************************************/
/* Fetch: reads a record from DXF file. Returns group / -1 on EOF.  */
/********************************************************************/

int CDxfin::Fetch()
{
	used = FALSE;
	addr = ftell(fp);
	line = lcnt;
	if(binary)
		FetchBin();
	else
		FetchAsc();
	return group;
}

/********************************************************************/
/* Record: reads record when used. Returns same as Fetch.           */
/********************************************************************/

int CDxfin::Record()
{
	if(used)
		CDxfin::Fetch();
	return group;
}

/********************************************************************/
/* Seek: seeks record with given group number. Returns FALSE on EOF.*/
/********************************************************************/

BOOL CDxfin::Seek(int g)
{
	while(Record() >= 0 && group != g)
		Use();
	return group == g;
}

/********************************************************************/
/* Str: seeks string record. Marks as Used. Returns FALSE on EOF.   */
/********************************************************************/

BOOL CDxfin::Str(int g, const char *s)
{
	while(Seek(g))
	{
		Use();
		if(!strcmp(strVal, s))
			return TRUE;
	}
	return FALSE;
}

/********************************************************************/
/* Section: compares group to 0 and value to SECTION.               */
/********************************************************************/

BOOL CDxfin::Section()
{
	return group == 0 && !strcmp(strVal, "SECTION");
}

/********************************************************************/
/* EndSec: compares group to 0 and value to ENDSEC.                 */
/********************************************************************/

BOOL CDxfin::EndSec()
{
	return group == 0 && !strcmp(strVal, "ENDSEC");
}

/********************************************************************/
/* GetPos: stores DXF position to a structure.                      */
/********************************************************************/

void CDxfin::GetPos(TDxfLabel &lbl)
{
	lbl.addr = addr;
	lbl.line = line;
	lbl.used = used;
}

/********************************************************************/
/* SetPos: moves DXF pointer to given position.                     */
/********************************************************************/

void CDxfin::SetPos(const TDxfLabel &lbl)
{
	fseek(fp, lbl.addr, SEEK_SET);
	lcnt = lbl.line;
	Fetch();
	used = lbl.used;
}

/********************************************************************/
/* ErrName: returns error description.                              */
/********************************************************************/

const char *CDxfin::ErrName()
{
	switch(group)
	{
	case DXF_EOF   : return "unexpected end of file";
	case DXF_FILE  : return "file failed to open";
	case DXF_FORMAT: return "unknown file format";
	case DXF_GROUP : return "group number expected";
	case DXF_INT   : return "integer value expected";
	case DXF_FLT   : return "real value expected";
	case DXF_LINE  : return "too long string/line";
	case DXF_EXTRA : return "extra characters on line";
	default        : return "syntax error";
	}
}

/********************************************************************/
/* ErrAddr: returns error address/line.                             */
/********************************************************************/

long CDxfin::ErrAddr()
{
	return binary ? addr : lcnt;
}

/********************************************************************/
/* Done: returns % of file processed.                               */
/********************************************************************/

int CDxfin::Done()
{
	long fpos = ftell(fp);

	if(size >= 1000000L)
		fpos = fpos / (size / 100);
	else
		fpos = fpos * 100 / size;
	return (int)fpos;
}

/********************************************************************/
/* PrintOn: displays current record on given stream.                */
/********************************************************************/

void CDxfin::PrintOn(FILE *stream)
{
	char *name = "";
	int   tpnm = "SIF?"[(unsigned)type < 3 ? type : 3];

	fprintf(stream, binary ? "[ %06x %c%c" : "[%5ld %c%c",
		(binary ? addr : line), used ? 'U' : 'u', tpnm);
	switch(group)
	{
	case DXF_EOF   : name = "EOF   "; break;
	case DXF_FILE  : name = "FILE  "; break;
	case DXF_FORMAT: name = "FORMAT"; break;
	case DXF_GROUP : name = "GROUP "; break;
	case DXF_INT   : name = "INT   "; break;
	case DXF_FLT   : name = "FLT   "; break;
	case DXF_LINE  : name = "LINE  "; break;
	case DXF_EXTRA : name = "EXTRA "; break;
	default        : fprintf(stream, "%6d", group); break;
	}
	fprintf(stream, "%s ", name);
	switch(type)
	{
	case RT_INT: fprintf(stream, "%d", intVal); break;
	case RT_FLT: fprintf(stream, "%g", fltVal); break;
	case RT_STR: fprintf(stream, "\"%s\"", strVal); break;
	default:     fprintf(stream, "(unknown type)"); break;
	}
	fprintf(stream, " ]\n");
}

/********************************************************************/
/* InitRead: initializes input; checks DXF type.                    */
/********************************************************************/

BOOL CDxfin::InitRead()
{
	int c1;

	size    = filelength(fileno(fp));
	binary  = FALSE;
	addr    = 0;
	lcnt    = 1;
	used    = TRUE;
	type    = RT_INT; /* to prevent error indication */
	group   = DXF_EOF;
	if((c1 = fgetc(fp)) == 'A')
	{ /* check binary type */
		char *tag = "AutoCAD Binary DXF\r\n\x1A";

		do
			if(*++tag != fgetc(fp))
				return FALSE;
		while(*tag);
		binary = TRUE;
	}
	else
	{
		for(; !isdigit(c1); c1 = fgetc(fp))
			if(c1 != ' ') /* invalid file */
				return FALSE;
		ungetc(c1, fp);
	}
	Fetch();
	return group == 0; /* 1st group must be class #0 */
}

/********************************************************************/
/* FetchAsc: reads a record from ASCII DXF.                         */
/********************************************************************/

void CDxfin::FetchAsc()
{
	group = DXF_EOF;
	do
	{
		for(int c1 = fgetc(fp); !isdigit(c1) && c1 >= 0; c1 = fgetc(fp))
			if(c1 != ' ')
			{ /* invalid group code */
				group = DXF_GROUP;
				return;
			}

		for(int code = 0; isdigit(c1); c1 = fgetc(fp))
			code = code * 10 + (c1 - '0');
		if(!SkipEol(c1))
			return; /* end of file / error */

		/* fetch record value */
		if((group = code) >= 999) /* make a -1 group out of comment */
			if((code -= 1000) < 0)
				group = -1; /* comment */

		if(code >= 60 && code < 80)
		{ /* integer value */
			BOOL neg = FALSE;

			while((c1 = fgetc(fp)) == ' ')
				;
			if(c1 == '+' || c1 == '-')
			{
				if(c1 == '-')
					neg = TRUE;
				c1 = fgetc(fp);
			}
			if(!isdigit(c1))
			{
				group = DXF_INT;
				return; /* invalid number */
			}
			lngVal = 0;
			do
			{
				lngVal = 10 * lngVal + (c1 - '0');
				c1 = fgetc(fp);
			}
			while(isdigit(c1));
			if(neg)
				lngVal = -lngVal;
			intVal = (int)lngVal;
			type = (group == 1071 ? RT_LNG : RT_INT);
		}
		else
		{ /* string or real value */
			char *ptr  = strVal;
			int   free = MAX_STRING;

			while((c1 = fgetc(fp)) >= ' ')
				if(--free)
					*ptr++ = c1;
				else
				{
					group = DXF_LINE;
					return; /* invalid (too long) line */
				}
			*ptr = 0;
			type = RT_STR;
			if(code >= 10)
			{ /* real value */
				errno = 0;
				fltVal = strtod(strVal, &ptr);
				if(errno)
				{
					group = DXF_FLT;
					return; /* invalid real number */
				}
				while(*ptr)
					if(*ptr++ != ' ')
					{ /* extra characters on line */
						group = DXF_EXTRA;
						return;
					}
				type = RT_FLT;
			}
		}
		if(!SkipEol(c1))
			return;
	}
	while(group < 0);
}

/********************************************************************/
/* FetchBin: reads a record from binary DXF.                        */
/********************************************************************/

void CDxfin::FetchBin()
{
	int code;

	group = DXF_EOF;
	if((code = fgetc(fp)) < 0)
		return; /* EOF */
	if(code == 0xFF)
	{ /* extended */
		code = getw(fp);
		if(feof(fp))
			return; /* EOF */
	}
	group = code;
	if(code == 1071)
	{ /* exception: LONG integer */
#if REVERSED_BYTE_ORDER
		int i = sizeof(long);
		do
			((char *)&lngVal)[--i] = fgetc(fp);
		while(i);
#else
		fread(&lngVal, sizeof(long), 1, fp);
#endif
		intVal = (int)lngVal;
		type = RT_LNG;
	}
	else
	{
		if(code >= 1000)
			code -= 1000;
		if(code < 10)
		{ /* string value */
			char *ptr  = strVal;
			int   free = MAX_STRING;

			for(int c1; (c1 = fgetc(fp)) > 0; *ptr++ = c1)
				if(!--free)
				{ /* too long string */
					group = DXF_LINE;
					return;
				}
			*ptr = 0;
			type = RT_STR;
		}
		else if(code >= 60 && code < 80)
		{
			intVal = getw(fp);
			type = RT_INT;
		}
		else
		{
#if REVERSED_BYTE_ORDER
			int i = sizeof(long);
			do
				((char *)&fltVal)[--i] = fgetc(fp);
			while(i);
#else
			fread(&fltVal, sizeof(double), 1, fp);
#endif
			type = RT_FLT;
		}
	}
	if(feof(fp)) /* eof */
		group = DXF_EOF;
}

/********************************************************************/
/* SkipEol: skips end of line, increments lcnt. FALSE=error/eof.    */
/********************************************************************/

BOOL CDxfin::SkipEol(int c1)
{
	for(; c1 != '\n'; c1 = fgetc(fp))
		if(c1 > ' ')
		{
			group = DXF_EXTRA;
			return FALSE;
		}
		else if(c1 < 0)
		{
			group = DXF_EOF;
			return FALSE;
		}
	lcnt++;
	return TRUE;
}
