/********************************************************************/
/* DXFREAD: high-level DXF input.                                   */
/********************************************************************/

#include <dxfread.h>
#define DIR_CHAR '\\'

#include <dynamem.h>
#include <math.h>

/*************************** C_StyleList:: **************************/

/**************************** constructor ***************************/

C_StyleList::C_StyleList(C_StyleList *cdr, CShxDraw *_shx, const char *_name)
{
	link = cdr;
	shx  = _shx;
	strcpy(name, _name);
}

/****************************** methods *****************************/

/********************************************************************/
/* Style: finds style with given name. Returns NULL on error.       */
/********************************************************************/

CShxDraw *C_StyleList::Style(const char *name)
{
	for(C_StyleList *list = this; list; list = list -> link)
		if(!strcmp(list -> name, name))
			return list -> shx;
	return NULL;
}

/********************************************************************/
/* Delete: deletes the whole list.                                  */
/********************************************************************/

void C_StyleList::Delete()
{
	for(C_StyleList *list = this, *next; list; list = next)
	{
		next = list -> link;
		delete list;
	}
}

/**************************** CPathList:: ***************************/

/****************************** methods *****************************/

/********************************************************************/
/* Make: creates a new path list entry.                             */
/********************************************************************/

CPathList *CPathList::Make(CPathList *join, const char far *_path)
{
	CPathList *ptr = new(_fstrlen(_path)) CPathList;
	ptr -> link = join;
	_fstrcpy(ptr -> path, _path);
	return ptr;
}

/*************************** C_BlockList:: **************************/

/**************************** constructor ***************************/

C_BlockList::C_BlockList(C_BlockList *cdr, TDxfLabel _addr, const char *_name)
: ins()
{
	link = cdr;
	addr = _addr;
	strcpy(name, _name);
}

/**************************** CDxfStack:: ***************************/

/**************************** constructor ***************************/

CDxfStack::CDxfStack(CDxfStack *join)
{
	link      = join;
	block     = NULL;
	orig      = CMatrix::E;
	row       = 0;
	row_count = 1;
	col       = 0;
	col_count = 1;
	angle     = 0;
	ins       = FP2D::O;
	spacing   = FP2D::O;
	scale     = FP2D(1, 1);
}

/****************************** methods *****************************/

/********************************************************************/
/* Matrix: computes combined block transformation matrix.           */
/********************************************************************/

CMatrix CDxfStack::Matrix()
{
	return CMatrix::Move(FP2D(-block -> ins))
	* CMatrix(FP2D(scale.x, 0), FP2D(0, scale.x))
	* CMatrix::Rotate(FP2D::O, angle)
	* CMatrix::Move(FP2D(ins.x + col * spacing.x, ins.y + row * spacing.y))
	* orig;
}

/*************************** CDxfReader:: ***************************/

/**************************** constructor ***************************/

CDxfReader::CDxfReader() : view(CNilView::O, CMatrix::E)
{
	blocks       = NULL;
	stack        = NULL;
	styles       = NULL;
	paths        = NULL;
	transform    = CMatrix::E;
	scanned      = FALSE;
	raw_text     = FALSE;
}

/**************************** destructor ****************************/

CDxfReader::~CDxfReader()
{
	Close();
}

/****************************** methods *****************************/

/********************************************************************/
/* Scan: scans file, makes block list.                              */
/********************************************************************/

BOOL CDxfReader::Scan()
{
	view.Matrix(transform);
	KillStack();
	if(scanned)
	{
		SetPos(entities);
		return TRUE;
	}
	scanned = TRUE;
	while(Str(0, "SECTION"))
		if(Record() == 2)
		{
			Use();
			switch(strVal[0])
			{
			case 'E': // check ENTITIES
				if(!strcmp(strVal, "ENTITIES"))
				{ // entities found; store pos & exit
					GetPos(entities);
					return TRUE;
				}
				break;

			case 'B': // check BLOCKS
				if(!strcmp(strVal, "BLOCKS"))
				{ // process BLOCK section
					while(Seek(0) && !EndSec())
					{
						Use();
						if(!strcmp(strVal, "BLOCK"))
						{ // store block addr, skip data
							TDxfLabel    addr;
							BOOL         name = FALSE;
							FP2D         ins;

							for(GetPos(addr); Record(); Use())
								switch(group)
								{
								case 2:
									if(name)
										return FALSE; // duplicate block name
									blocks = new(strVal)
									C_BlockList(blocks, addr, strVal);
									name = TRUE;
									break;

								case 10: ins.x = fltVal; break;
								case 20: ins.y = fltVal; break;
								}

							if(name == FALSE)
								return FALSE; // missing block name
							blocks -> ins = ins;
						}
					}
				}
				break;
			}
		}
	return FALSE;
}

/********************************************************************/
/* StackPop: removes 1 item off the top of the stack.               */
/********************************************************************/

void CDxfReader::StackPop()
{
	CDxfStack *tos;

	if((tos = stack) != NULL)
	{
		view.Matrix(tos -> orig);
		stack = tos -> link;
		if(fp)
			SetPos(tos -> addr);
		delete tos;
	}
}

/********************************************************************/
/* StackPush: adds 1 item to the top of the stack.                  */
/********************************************************************/

void CDxfReader::StackPush()
{
	stack = new CDxfStack(stack);
	stack -> orig = view.Matrix();
	GetPos(stack -> addr);
}

/********************************************************************/
/* KillStack: deallocates stack.                                    */
/********************************************************************/

void CDxfReader::KillStack()
{
	while(stack)
		StackPop();
}

/********************************************************************/
/* Close: closes file, deallocates lists.                           */
/********************************************************************/

void CDxfReader::Close()
{
	CDxfin::Close();
	scanned = FALSE;
	KillStack();
	styles -> Delete();
	styles = NULL;
	for(C_BlockList *list; blocks; blocks = list)
	{
		list = blocks -> link;
		delete blocks;
	}
}

/********************************************************************/
/* AddPath: adds a path to path list.                               */
/********************************************************************/

void CDxfReader::AddPath(const char far *path)
{
	paths = CPathList::Make(paths, path);
}

/********************************************************************/
/* Style: finds given style. Loads font when necessary.             */
/********************************************************************/

CShxDraw *CDxfReader::Style(const char *n)
{
	CShxDraw *sty = styles -> Style(n);

	if(sty == NULL)
	{ // load style
		char  name[FILENAME_MAX];

		for(CPathList *plist = paths; plist; plist = plist -> link)
		{
			int len = strlen(strcpy(name, plist -> path));

			if(len && name[len - 1] != DIR_CHAR)
				name[len++] = DIR_CHAR;
			strcpy(name + len, n);
			strcat(name, ".shx");
			sty = new CShxDraw(name);
			if(sty -> IsValid())
			{ // valid style - add to style table
				styles = new(strlen(n)) C_StyleList(styles, sty, n);
				return sty;
			}
			delete sty;
		}
	}
	return sty;
}


/********************************************************************/
/* OnEntity: returns FALSE in this class.                           */
/********************************************************************/

BOOL CDxfReader::OnEntity()
{
	return FALSE;
}

/********************************************************************/
/* ExpandText: expands text entity.                                 */
/********************************************************************/

void CDxfReader::ExpandText()
{
	static CJoint vert(FP2D(0, 0), FP2D(0, 1));
	static CJoint horz(FP2D(0, 0), FP2D(1, 0));

	char         text[CDxfin::MAX_STRING];
	char         style[CDxfin::MAX_STRING];
	FP2D         ins;
	FP2D         align = FP2D::NIL;
	double       height = 1;
	double       angle  = 0;
	double       pitch  = 0;
	double       width  = 1;
	int          mirror = 0;
	int          justif = 0;
	CShxDraw    *shx;

	strcpy(style, "STANDARD");
	for(*text = 0; Record(); Use())
		switch(group)
		{
		case 1: strcpy(text, strVal); break;
		case 7: strcpy(style, strVal); break;

		case 10: ins.x = fltVal; break;
		case 20: ins.y = fltVal; break;

		case 11: align.x = fltVal; break;
		case 21: align.y = fltVal; break;

		case 40: height = fltVal; break;
		case 41: width  = fltVal; break;

		case 50: angle  = fltVal * (M_PI / 180); break;
		case 51: pitch  = fltVal * (M_PI / 180); break;

		case 71: mirror = intVal; break;
		case 72:
			switch(intVal)
			{
			default:
			case 0: justif = 000; break;
			case 1: justif = 001; break;
			case 2: justif = 002; break;
			case 3: justif = 011; break;
			case 4: justif = 011; break;
			case 5: justif = 011; break;
			}
		}

	if(align.IsNull())
		align = ins;

	if((shx = Style(style)) != NULL)
	{
		CMatrix form = CMatrix::Scale(FP2D::O, FP2D(height * width, height));
		if(raw_text)
			form = form * CMatrix::Move(align);
		else
		{
			form = form * CMatrix::Pitch(pitch);
			if(mirror & 2)
				form = form * CMatrix::Mirror(vert);
			if(mirror & 4)
				form = form * CMatrix::Mirror(horz);
			form = form * CMatrix::Rotate(FP2D::O, angle) * CMatrix::Move(ins);
		}
		form = form * view.Matrix();
		CTransformedView formed(view.View(), form);

		if(raw_text)
		{ // measure text and send it right through to viewport
			CLimiterView limiter;

			// measure text extents
			shx -> OnView(&limiter);
			shx -> Expand(text);

			// compute text generation parameters
			double wtext = limiter.limits.Width();
			double htext = 1;

			formed.Text(FP2D::O, wtext, htext, angle, pitch, justif, text);
		}
		else
		{ // compute text output matrix & render text
			shx -> OnView(&formed);
			formed.MoveTo(FP2D::O);
			shx -> Expand(text);
		}
	}
}

/********************************************************************/
/* ExpandCircle: performs CIRCLE command.                           */
/********************************************************************/

void CDxfReader::ExpandCircle()
{
	FP2D   ctr;
	double rad = 0;

	for(; Record(); Use())
		switch(group)
		{
		case 10: ctr.x = fltVal; break;
		case 20: ctr.y = fltVal; break;
		case 40: rad   = fltVal; break;
		}

	view.Arc(CBulge(ctr, rad, 0, 2 * M_PI));
}

/********************************************************************/
/* ExpandArc: expands arc entity.                                   */
/********************************************************************/

void CDxfReader::ExpandArc()
{
	FP2D   ctr;
	double rad   = 0;
	double alpha = 0;
	double beta  = 0;

	for(; Record(); Use())
		switch(group)
		{
		case 10: ctr.x = fltVal; break;
		case 20: ctr.y = fltVal; break;
		case 40: rad   = fltVal; break;
		case 50: alpha = fltVal * (M_PI / 180); break;
		case 51: beta  = fltVal * (M_PI / 180); break;
		}

	if(view.Matrix().Det() < 0) // clockwise
		while(alpha <= beta)
			beta -= 2 * M_PI;
	else
		while(alpha >= beta)
			beta += 2 * M_PI;

	view.Arc(CBulge(ctr, rad, alpha, beta));
}

/********************************************************************/
/* ExpandPoint: reads & draws a point entity.                       */
/********************************************************************/

void CDxfReader::ExpandPoint()
{
	FP2D P;

	for(; Record(); Use())
		switch(group)
		{
		case 10: P.x = fltVal; break;
		case 20: P.y = fltVal; break;
		}

	view.Point(P);
}

/********************************************************************/
/* ExpandLine: reads & draws a line entity.                         */
/********************************************************************/

void CDxfReader::ExpandLine()
{
	FP2D P, Q;

	for(; Record(); Use())
		switch(group)
		{
		case 10: P.x = fltVal; break;
		case 20: P.y = fltVal; break;

		case 11: Q.x = fltVal; break;
		case 21: Q.y = fltVal; break;
		}

	view.Line(P, Q);
}

/********************************************************************/
/* ExpandPolyline: reads & draws a polyline entity.                 */
/********************************************************************/

void CDxfReader::ExpandPolyline()
{
	int    flags = 0;
	int    count = 0;
	FP2D   vclose;
	double bulge = 0;

	for(; Record(); Use())
		switch(group)
		{
		case 70: flags = intVal; break;
		}

	while(Seek(0) && !strcmp(strVal, "VERTEX"))
	{
		FP2D   vertex;
		double new_bulge = 0;
		int    flags = 0;

		for(Use(); Record(); Use())
			switch(group)
			{
			case 10: vertex.x  = fltVal; break;
			case 20: vertex.y  = fltVal; break;
			case 42: new_bulge = fltVal; break;
			case 70: flags     = intVal; break;
			}

		if(flags & 16)
			continue; // ignore spline control points

		switch(count)
		{
		case 0:
			vclose = vertex;
			view.MoveTo(vertex);
			count++;
			break;

		case 1:
		case 2:
			count++;
		default:
			if(bulge)
				view.Arc(CBulge(view.cursor, vertex,
				bulge * (view.cursor | vertex) / 2));
			else
				view.LineTo(vertex);
			break;
		}
		bulge = new_bulge;
	}

	if(count > 2 && (flags & 1) != 0) // close poly
		if(bulge)
			view.Arc(CBulge(view.cursor, vclose,
			bulge * (view.cursor | vclose) / 2));
		else
			view.LineTo(vclose);

	return;
}

/********************************************************************/
/* Expand: processes file as linear; calls OnEntity for each entity.*/
/********************************************************************/

BOOL CDxfReader::Expand(void (*progress)(int))
{
	int step = 0;

	if(progress)
		progress(0);
	if(!Scan())
		return FALSE;
	while(Seek(0) && !EndSec())
	{
		if(progress && --step < 0)
		{ // show progress
			int done = Done();
			progress(done <= 0 ? 1 : (done >= 100 ? 99 : done));
			step = 100;
		}

		Use();
		if(!OnEntity())
			switch(strVal[0])
			{ // expand ignored cases
			case 'A': // arc
				if(!strcmp(strVal, "ARC")) // expand arc
					ExpandArc();
				break;

			case 'C': // circle
				if(!strcmp(strVal, "CIRCLE")) // expand circle
					ExpandCircle();
				break;

			case 'E': // endblk
				if(!strcmp(strVal, "ENDBLK")) // return from block
					LeaveBlock();
				break;

			case 'I': // insert
				if(!strcmp(strVal, "INSERT")) // nest into block
					EnterBlock();
				break;

			case 'L': // line
				if(!strcmp(strVal, "LINE"))
					ExpandLine();
				break;

			case 'P': // polyline
				if(!strcmp(strVal, "POLYLINE"))
					ExpandPolyline();
				else if(!strcmp(strVal, "POINT"))
					ExpandPoint();
				break;

			case 'S': // later - expand shapes
				break;

			case 'T':
				if(!strcmp(strVal, "TEXT"))
					ExpandText();
				break;
			}
	}
	if(progress)
		progress(100);
	return TRUE;
}

/********************************************************************/
/* EnterBlock: reads INSERT entity and goes into the block.         */
/********************************************************************/

void CDxfReader::EnterBlock()
{
	StackPush();
	for(; Record(); Use())
		switch(group)
		{
		case 2: stack -> block = Block(strVal); break;

		case 10: stack -> ins.x = fltVal; break;
		case 20: stack -> ins.y = fltVal; break;

		case 41: stack -> scale.x = fltVal; break;
		case 42: stack -> scale.y = fltVal; break;

		case 44: stack -> spacing.x = fltVal; break;
		case 45: stack -> spacing.y = fltVal; break;

		case 50: stack -> angle = fltVal; break;

		case 70: stack -> col_count = intVal; break;
		case 71: stack -> row_count = intVal; break;
		}

	if(stack -> block == NULL)
	{
		StackPop();
		return; // invalid name or block not found
	}

	view.Matrix(stack -> Matrix());
	SetPos(stack -> block -> addr);
}

/********************************************************************/
/* LeaveBlock: repeats block if multiple, leaves if not.            */
/********************************************************************/

void CDxfReader::LeaveBlock()
{
	if(!stack)
		return;
	if(++stack -> col >= stack -> col_count)
	{
		stack -> col = 0;
		if(++stack -> row >= stack -> row_count)
		{ // end of block - quit
			StackPop();
			return;
		}
	}
	view.Matrix(stack -> Matrix());
	SetPos(stack -> block -> addr);
}

/********************************************************************/
/* Block: finds block with given name.                              */
/********************************************************************/

C_BlockList *CDxfReader::Block(const char *name)
{
	for(C_BlockList *list = blocks; list; list = list -> link)
		if(!strcmp(list -> name, name))
			break;
	return list;
}
