/********************************************************************/
/* FP: real coordinate 2D & 3D vector.                              */
/********************************************************************/

#include <fp.h>
#include <math.h>

const double NIL = DBL_MAX;

/********************************************************************/
/* ftoi: truncates a real number to integer range.                  */
/********************************************************************/

int ftoi(double x)
{
	return (x <= -32767 ? -32767 : x >= 32767 ? 32767 : (int)x);
}

/****************************** FP2D :: *****************************/

/**************************** static data ***************************/

FP2D FP2D::O;
FP2D FP2D::NIL(::NIL, ::NIL);

/*********************** typecasting operators **********************/

/********************************************************************/
/* operator POINT: converts point to integer point.         */
/********************************************************************/

FP2D::operator POINT () const
{
	POINT tmp;
	
	tmp.x = ftoi(x);
	tmp.y = ftoi(y);

	return tmp;
}

/********************************************************************/
/* operator LONG: converts point to longword point.                 */
/********************************************************************/

FP2D::operator LONG () const
{
	POINT tmp = (POINT)*this;

	return MAKELONG(tmp.x, tmp.y);
}

/********************************************************************/
/* operator + : adds 2 points.                                      */
/********************************************************************/

FP2D FP2D::operator + (const FP2D &q) const
{
	return FP2D(x + q.x, y + q.y);
}

/********************************************************************/
/* operator - : substracts 2 points.                                */
/********************************************************************/

FP2D FP2D::operator - (const FP2D &q) const
{
	return FP2D(x - q.x, y - q.y);
}

/********************************************************************/
/* operator * : multiplies a point by a ratio.                      */
/********************************************************************/

FP2D FP2D::operator * (double r) const
{
	return FP2D(x * r, y * r);
}

/********************************************************************/
/* operator / : divides a point by a ratio.                         */
/********************************************************************/

FP2D FP2D::operator / (double r) const
{
	return FP2D(x / r, y / r);
}

/********************************************************************/
/* operator += : adds a point to this point.                        */
/********************************************************************/

FP2D &FP2D::operator += (const FP2D &q)
{
	x += q.x;
	y += q.y;

	return *this;
}

/********************************************************************/
/* operator -= : substracts a point from this point.                */
/********************************************************************/

FP2D &FP2D::operator -= (const FP2D &q)
{
	x -= q.x;
	y -= q.y;

	return *this;
}

/********************************************************************/
/* operator *= : multiplies the point by a ratio.                   */
/********************************************************************/

FP2D &FP2D::operator *= (double r)
{
	x *= r;
	y *= r;

	return *this;
}

/********************************************************************/
/* operator /= : divides the point by a ratio.                      */
/********************************************************************/

FP2D &FP2D::operator /= (double r)
{
	x /= r;
	y /= r;

	return *this;
}

/********************************************************************/
/* operator <= : accumulates minimum of coordinates.                */
/********************************************************************/

FP2D &FP2D::operator <= (const FP2D &q)
{
	if(q.x < x)
		x = q.x;
	if(q.y < y)
		y = q.y;
	return *this;
}

/********************************************************************/
/* operator >= : accumulates maximum of coordinates.                */
/********************************************************************/

FP2D &FP2D::operator >= (const FP2D &q)
{
	if(q.x > x)
		x = q.x;
	if(q.y > y)
		y = q.y;
	return *this;
}

/********************************************************************/
/* operator * : computes scalar product of 2 points.                */
/********************************************************************/

double FP2D::operator * (const FP2D &q) const
{
	return x * q.x + y * q.y;
}

/********************************************************************/
/* operator % : computes determinant of 2 points.                   */
/********************************************************************/

double FP2D::operator % (const FP2D &q) const
{
	return x * q.y - y * q.x;
}

/********************************************************************/
/* operator | : computes distance of 2 points.                      */
/********************************************************************/

double FP2D::operator | (const FP2D &q) const
{
	return hypot(x - q.x, y - q.y);
}

/********************************************************************/
/* operator ^: returns projected length.                            */
/********************************************************************/

double FP2D::operator ^ (const FP2D &q) const
{
	double product = *this * q;
	double measure = q.length();

	return measure ? product / measure : 0;
}

/********************************************************************/
/* scale: scales a point to given distance from a centre.           */
/********************************************************************/

FP2D FP2D::scale(const FP2D &O, double to) const
{
	double dist = *this | O;

	if(dist > to / 1e6)
		to /= dist;
	return O + (*this - O) * to;
}

/********************************************************************/
/* rotate: rotates a point around another point.                    */
/********************************************************************/

FP2D FP2D::rotate(const FP2D &O, double ang) const
{
	double sa = sin(ang);
	double ca = cos(ang);
	FP2D D  = *this - O;

	return O + FP2D(D.x * ca - D.y * sa, D.x * sa + D.y * ca);
}

/********************************************************************/
/* unit: computes unit vector in given direction.                   */
/********************************************************************/

FP2D FP2D::unit() const
{
	return scale(O, 1.0);
}

/********************************************************************/
/* middle: computes the average of 2 points.                        */
/********************************************************************/

FP2D FP2D::middle(const FP2D &B) const
{
	return (*this + B) / 2;
}

/********************************************************************/
/* bearing: returns point bearing (angle in polar coords).          */
/********************************************************************/

double FP2D::bearing() const
{
	if(y == 0)
		return (x >= 0 ? 0 : M_PI);
	if(x == 0)
		return (y >= 0 ? M_PI_2 : 3 * M_PI_2);
	return atan2(y, x);
}

/********************************************************************/
/* squared: returns squared vector length.                          */
/********************************************************************/

double FP2D::squared() const
{
	return x * x + y * y;
}

/********************************************************************/
/* length: returns the length of this vector.                       */
/********************************************************************/

double FP2D::length() const
{
	return sqrt(squared());
}

/********************************************************************/
/* IsNull: checks whether point is NIL.                             */
/********************************************************************/

BOOL FP2D::IsNull() const
{
	return x == ::NIL || y == ::NIL;
}

/********************************************************************/
/* operator ==: compares 2 vectors.                                 */
/********************************************************************/

int FP2D::operator == (const FP2D &pt) const
{
	return x == pt.x && y == pt.y;
}

/********************************************************************/
/* Polar: constructs a point in polar coords.                       */
/********************************************************************/

FP2D FP2D::Polar(double r, double a)
{
	return FP2D(r * cos(a), r * sin(a));
}

/****************************** FP3D :: *****************************/

/**************************** static data ***************************/

FP3D FP3D::O;
FP3D FP3D::NIL(::NIL, ::NIL, ::NIL);

/*********************** typecasting operators **********************/

/********************************************************************/
/* operator + : adds 2 points.                                      */
/********************************************************************/

FP3D FP3D::operator + (const FP3D &q) const
{
	return FP3D(x + q.x, y + q.y, z + q.z);
}

/********************************************************************/
/* operator - : substracts 2 points.                                */
/********************************************************************/

FP3D FP3D::operator - (const FP3D &q) const
{
	return FP3D(x - q.x, y - q.y, z - q.z);
}

/********************************************************************/
/* operator * : multiplies a point by a ratio.                      */
/********************************************************************/

FP3D FP3D::operator * (double r) const
{
	return FP3D(x * r, y * r, z * r);
}

/********************************************************************/
/* operator / : divides a point by a ratio.                         */
/********************************************************************/

FP3D FP3D::operator / (double r) const
{
	return FP3D(x / r, y / r, z / r);
}

/********************************************************************/
/* operator += : adds a point to this point.                        */
/********************************************************************/

FP3D &FP3D::operator += (const FP3D &q)
{
	x += q.x;
	y += q.y;
	z += q.z;

	return *this;
}

/********************************************************************/
/* operator -= : substracts a point from this point.                */
/********************************************************************/

FP3D &FP3D::operator -= (const FP3D &q)
{
	x -= q.x;
	y -= q.y;
	z -= q.z;

	return *this;
}

/********************************************************************/
/* operator *= : multiplies the point by a ratio.                   */
/********************************************************************/

FP3D &FP3D::operator *= (double r)
{
	x *= r;
	y *= r;
	z *= r;

	return *this;
}

/********************************************************************/
/* operator /= : divides the point by a ratio.                      */
/********************************************************************/

FP3D &FP3D::operator /= (double r)
{
	x /= r;
	y /= r;
	z /= r;

	return *this;
}

/********************************************************************/
/* operator <= : accumulates minimum of coordinates.                */
/********************************************************************/

FP3D &FP3D::operator <= (const FP3D &q)
{
	if(q.x < x)
		x = q.x;
	if(q.y < y)
		y = q.y;
	if(q.z < z)
		z = q.z;
	return *this;
}

/********************************************************************/
/* operator >= : accumulates maximum of coordinates.                  */
/********************************************************************/

FP3D &FP3D::operator >= (const FP3D &q)
{
	if(q.x > x)
		x = q.x;
	if(q.y > y)
		y = q.y;
	if(q.z > z)
		z = q.z;
	return *this;
}

/********************************************************************/
/* operator * : computes scalar product of 2 points.                */
/********************************************************************/

double FP3D::operator * (const FP3D &q) const
{
	return x * q.x + y * q.y + z * q.z;
}

/********************************************************************/
/* operator % : computes vector product of 2 points.                */
/********************************************************************/

FP3D FP3D::operator % (const FP3D &q) const
{
	return FP3D(y * q.z - z * q.y, z * q.x - x * q.z, x * q.y - y * q.x);
}

/********************************************************************/
/* operator | : computes distance of 2 points.                      */
/********************************************************************/

double FP3D::operator | (const FP3D &q) const
{
	return (*this - q).length();
}

/********************************************************************/
/* operator ^: returns projected length.                            */
/********************************************************************/

double FP3D::operator ^ (const FP3D &q) const
{
	double product = *this * q;
	double measure = q.length();

	return measure ? product / measure : 0;
}

/********************************************************************/
/* mix: mixed product of 3 points - determinant.                    */
/********************************************************************/

double FP3D::mix(const FP3D &P, const FP3D &Q) const
{
	return
	+ x * (P.y * Q.z - P.z * Q.y)
	+ y * (P.z * Q.x - P.x * Q.z)
	+ z * (P.x * Q.y - P.y * Q.x);
}

/********************************************************************/
/* scale: scales a point to given distance from a centre.           */
/********************************************************************/

FP3D FP3D::scale(const FP3D &O, double to) const
{
	double dist = *this | O;

	if(dist > to / 1e6)
		to /= dist;
	return O + (*this - O) * to;
}

/********************************************************************/
/* unit: computes unit vector in given direction.                   */
/********************************************************************/

FP3D FP3D::unit() const
{
	return scale(O, 1.0);
}

/********************************************************************/
/* middle: computes the average of 2 points.                        */
/********************************************************************/

FP3D FP3D::middle(const FP3D &B) const
{
	return (*this + B) / 2;
}

/********************************************************************/
/* squared: returns squared vector length.                          */
/********************************************************************/

double FP3D::squared() const
{
	return x * x + y * y + z * z;
}

/********************************************************************/
/* length: returns the length of this vector.                       */
/********************************************************************/

double FP3D::length() const
{
	return sqrt(squared());
}

/********************************************************************/
/* IsNull: checks whether point is NIL.                             */
/********************************************************************/

BOOL FP3D::IsNull() const
{
	return x == ::NIL || y == ::NIL || z == ::NIL;
}

/********************************************************************/
/* operator ==: compares 2 vectors.                                 */
/********************************************************************/

int FP3D::operator == (const FP3D &pt) const
{
	return x == pt.x && y == pt.y && z == pt.z;
}
