/********************************************************************/
/* GEOMETRY: analytic geometry computations.                        */
/********************************************************************/

#include <geometry.h>
#include <math.h>

/******************************* data *******************************/

static const double EPSILON = 1e-10;

/****************************** CBox:: ******************************/

/**************************** static data ***************************/

CBox CBox::NIL;

/**************************** constructor ***************************/

CBox::CBox()
{
	LB = RT = FP2D::NIL;
}

CBox::CBox(const FP2D &P, const FP2D &Q)
{
	Set(P, Q);
}

/****************************** methods *****************************/

/********************************************************************/
/* Set: changes box limits.                                         */
/********************************************************************/

void CBox::Set(const FP2D &P, const FP2D &Q)
{
	LB  = P;
	LB <= Q;
	RT  = P;
	RT >= Q;
}

/********************************************************************/
/* Contains: answers TRUE when rectangle contains given point.      */
/********************************************************************/

BOOL CBox::Contains(const FP2D &P) const
{
	if(IsNull())
		return FALSE;
	return (P.x >= LB.x && P.x <= RT.x && P.y >= LB.y && P.y <= RT.y);
}

/********************************************************************/
/* operator +=: includes a point into the rectangle.                */
/********************************************************************/

void CBox::operator += (const FP2D &P)
{
	if(IsNull())
		LB = RT = P;
	else
	{
		LB <= P;
		RT >= P;
	}
}

/********************************************************************/
/* IsNull: checks whether the rectangle is empty.                   */
/********************************************************************/

BOOL CBox::IsNull() const
{
	return LB.IsNull() || RT.IsNull();
}

/***************************** CJoint:: *****************************/

/**************************** static data ***************************/

CJoint CJoint::NIL;

/**************************** constructor ***************************/

CJoint::CJoint()
{
	A = B = FP2D::NIL;
}

/****************************** methods *****************************/

/********************************************************************/
/* Vector: return the vector from start point to end point.         */
/********************************************************************/

FP2D CJoint::Vector() const
{
	return B - A;
}

/********************************************************************/
/* Vector: sets end point to start point plus given vector.         */
/********************************************************************/

void CJoint::Vector(const FP2D &V)
{
	B = A + V;
}

/********************************************************************/
/* AtY: computes x at given y coord. NIL if out of line.            */
/********************************************************************/

double CJoint::AtY(double y) const
{
	if(y > A.y && y > B.y || y < A.y && y < B.y)
		return ::NIL;
	if(y == A.y)
		return A.x;
	if(y == B.y)
		return B.x;
	return A.x + (B.x - A.x) * (y - A.y) / (B.y - A.y);
}

/********************************************************************/
/* Centre: returns line centre.                                     */
/********************************************************************/

FP2D CJoint::Centre() const
{
	return A.middle(B);
}

/********************************************************************/
/* In: true when the whole line is within a rectangle.              */
/********************************************************************/

BOOL CJoint::In(const CBox &box) const
{
	return box.Contains(A) && box.Contains(B);
}

/********************************************************************/
/* Intersect: computes intersection with another line.              */
/********************************************************************/

FP2D CJoint::Intersect(const CJoint &L) const
{
	FP2D   AB  = Vector();
	FP2D   CD  = L.Vector();
	FP2D   AC  = L.A - A;
	double det = AB % CD;
	double dt  = AC % CD;
	double du  = AC % AB;

	if(det != 0)
	{
		if(det < 0)
		{
			det = -det;
			dt  = -dt;
			du  = -du;
		}
		if(dt >= 0 && dt <= det && du >= 0 && du <= det)
			return A + AB * (dt / det);
	}
	return FP2D::NIL;
}

/********************************************************************/
/* Length: returns line length.                                     */
/********************************************************************/

double CJoint::Length() const
{
	return A | B;
}

/********************************************************************/
/* Clip: clips off all parts out of the rectangle.                  */
/********************************************************************/

CJoint CJoint::Clip(const CBox &box) const
{
	FP2D P, Q;

	if(IsNull() || box.IsNull())
		return NIL;

	if(A.x <= B.x)
		P = A, Q = B;
	else
		P = B, Q = A;

	/* now clip x coordinate; P.x <= Q.x */
	if(Q.x <= box.LB.x || P.x >= box.RT.x)
		return NIL;

	if(P.x < box.LB.x)
	{
		P.y += (Q.y - P.y) * (P.x - box.LB.x) / (P.x - Q.x);
		P.x  = box.LB.x;
	}
	if(Q.x > box.RT.x)
	{
		Q.y += (P.y - Q.y) * (Q.x - box.RT.x) / (Q.x - P.x);
		Q.x  = box.RT.x;
	}

	if(P.y > Q.y)
	{
		FP2D T;
		T = P, P = Q, Q = T;
	}

	/* now clip y coordinate. P.y <= P.y. */
	if(Q.y <= box.LB.y || P.y >= box.RT.y)
		return NIL;
	if(P.y < box.LB.y)
	{
		P.x += (Q.x - P.x) * (P.y - box.LB.y) / (P.y - Q.y);
		P.y  = box.LB.y;
	}
	if(Q.y > box.RT.y)
	{
		Q.x += (P.x - Q.x) * (Q.y - box.RT.y) / (Q.y - P.y);
		Q.y  = box.RT.y;
	}

	if(P == Q)
		return NIL;

	return CJoint(P, Q);
}

/********************************************************************/
/* IsNull: checks whether a line is null.                           */
/********************************************************************/

BOOL CJoint::IsNull() const
{
	return A.IsNull() || B.IsNull();
}

/********************************************************************/
/* PrintOn: prints line on a stream.                                */
/********************************************************************/

void CJoint::PrintOn(FILE *stream) const
{
	fprintf(stream, "(%g,%g -> %g,%g)", A.x, A.y, B.x, B.y);
}

/***************************** CBulge:: *****************************/

/****************************** methods *****************************/

/********************************************************************/
/* Init: initializes the arc with a bow and height.                 */
/********************************************************************/

void CBulge::Init(const FP2D &X, const FP2D &Y, double height)
{
	CJoint joint(X, Y);
	Init(joint, joint.Centre() + joint.Vector().right().scale(FP2D::O, height));
}

/********************************************************************/
/* Init: initializes the arc with a bow and point on arc.           */
/********************************************************************/

void CBulge::Init(const CJoint &_bow, const FP2D &_point)
{
	if(_bow.Length() < EPSILON)
	{ // singular bow: A = B
		isArc = FALSE;
		return;
	}

	double x2  = _point.squared();
	FP2D   AX  = _point - _bow.A;
	FP2D   BX  = _point - _bow.B;
	double m1x = (x2 - _bow.A.squared()) / 2;
	double m1y = (x2 - _bow.B.squared()) / 2;
	double det = AX % BX;
	FP2D   dc(m1x * BX.y - m1y * AX.y, m1y * AX.x - m1x * BX.x);

	bow        = _bow;
	pointOnArc = _point;

	if(det == 0)
		isArc = FALSE;
	else
	{
		isArc = TRUE;

		centre    = dc / det;
		radius    = pointOnArc | centre;
		alpha     = (_bow.A - centre).bearing();
		beta      = (_bow.B - centre).bearing();

		if(det < 0)
		{
			if(alpha > beta)
				beta += 2 * M_PI;
		}
		else
		{
			if(beta > alpha)
				alpha += 2 * M_PI;
		}
		height = (det >= 0 ? -radius : +radius)
		- _bow.Vector().unit() % (centre - bow.Centre());
	}
}

/********************************************************************/
/* Init: initializes the arc using centre, radius & angles.         */
/********************************************************************/

void CBulge::Init(const FP2D &C, double r, double _alpha, double _beta)
{
	centre     = C;
	radius     = r;
	alpha      = _alpha;
	beta       = _beta;
	bow.A      = C + FP2D::Polar(r, alpha);
	bow.B      = C + FP2D::Polar(r, beta);
	pointOnArc = C + FP2D::Polar(r, (alpha + beta) / 2.0);
	height     = bow.Centre() | pointOnArc;
	if(beta < alpha)
		height = -height;
	isArc = (radius != 0 && alpha != beta);
}

/********************************************************************/
/* Init: initializes a degenerated line arc.                        */
/********************************************************************/

void CBulge::Init(const CJoint &_bow)
{
	bow        = _bow;
	pointOnArc = _bow.Centre();
	height     = 0;
	isArc      = FALSE;
}

/********************************************************************/
/* PrintOn: prints the arc on an output stream.                     */
/********************************************************************/

void CBulge::PrintOn(FILE *stream) const
{
	fprintf(stream, "Arc bow:      "); bow.PrintOn(stream); fputc('\n', stream);
	fprintf(stream, "Point on arc: [%g,%g]\n", pointOnArc.x, pointOnArc.y);
	if(isArc)
	{
		fprintf(stream, "Centre: [%g,%g]\n", centre.x, centre.y);
		fprintf(stream, "Radius: %g\n", radius);
		fprintf(stream, "Height: %g\n", height);
		fprintf(stream, "Alpha:  %g\n", alpha);
		fprintf(stream, "Beta:   %g\n", beta);
	}
	else
		fprintf(stream, "(no bulge)\n");
}

/***************************** CMatrix:: ****************************/

/**************************** static data ***************************/

CMatrix CMatrix::E(FP2D(1, 0), FP2D(0, 1));

/**************************** constructor ***************************/

CMatrix::CMatrix()
{
	vx.x = vy.y = 1;
	vx.y = vy.x = 0;
	va.x = va.y = 0;
}

CMatrix::CMatrix(const FP2D &X, const FP2D &Y)
{
	vx = X;
	vy = Y;
	va.x = va.y = 0;
}

CMatrix::CMatrix(const FP2D &X, const FP2D &Y, const FP2D &A)
{
	vx = X;
	vy = Y;
	va = A;
}

/****************************** methods *****************************/

/********************************************************************/
/* Mirror: creates a mirroring matrix.                              */
/********************************************************************/

CMatrix CMatrix::Mirror(const CJoint &line)
{
	FP2D    V  = line.Vector();
	double  v2 = V.squared();
	double  a, b;

	if(!v2)
		return E;

	a = (V.x * V.x - V.y * V.y) / v2;
	b = 2 * V.x * V.y / v2;
	return CMatrix(FP2D(a, b), FP2D(b, -a)).Keep(line.A);
}

/********************************************************************/
/* Move: creates a displacement matrix.                             */
/********************************************************************/

CMatrix CMatrix::Move(const FP2D &V)
{
	CMatrix M;

	M.va = V;
	return M;
}

/********************************************************************/
/* Rotate: creates a revolution matrix.                             */
/********************************************************************/

CMatrix CMatrix::Rotate(const FP2D &C, double ang)
{
	double sa = sin(ang);
	double ca = cos(ang);

	return CMatrix(FP2D(ca, sa), FP2D(-sa, ca)).Keep(C);
}

/********************************************************************/
/* Scale: creates a scaling matrix.                                 */
/********************************************************************/

CMatrix CMatrix::Scale(const FP2D &C, double t)
{
	return CMatrix(FP2D(t, 0), FP2D(0, t)).Keep(C);
}

/********************************************************************/
/* Scale: creates an anisotropic scaling matrix.                    */
/********************************************************************/

CMatrix CMatrix::Scale(const FP2D &C, const FP2D &S)
{
	return CMatrix(FP2D(S.x, 0), FP2D(0, S.y)).Keep(C);
}

/********************************************************************/
/* Pitch: creates a pitch matrix.                                   */
/********************************************************************/

CMatrix CMatrix::Pitch(double a)
{
	return CMatrix(FP2D(1, 0), FP2D(tan(a), 1));
}

/********************************************************************/
/* Keep: changes va so that M * X = X.                              */
/********************************************************************/

CMatrix &CMatrix::Keep(const FP2D &X)
{
	va = X - X % (*this);
	return *this;
}

/********************************************************************/
/* operator *: multiplies the matrix by a point.                    */
/********************************************************************/

FP2D operator * (const FP2D &P, const CMatrix &M)
{
	return P % M + M.va;
}

/********************************************************************/
/* operator *: transforms a line through the matrix.                */
/********************************************************************/

CJoint operator * (const CJoint &J, const CMatrix &M)
{
	return CJoint(J.A * M, J.B * M);
}

/********************************************************************/
/* operator *: multiplies 2 matrices.                               */
/********************************************************************/

CMatrix operator * (const CMatrix &P, const CMatrix &Q)
{
	return CMatrix(
	FP2D(Q.vx.x * P.vx.x + Q.vy.x * P.vx.y, Q.vx.y * P.vx.x + Q.vy.y * P.vx.y),
	FP2D(Q.vx.x * P.vy.x + Q.vy.x * P.vy.y, Q.vx.y * P.vy.x + Q.vy.y * P.vy.y),
	P.va * Q);
}

/********************************************************************/
/* operator *: multiplies the bulge by a matrix.                    */
/********************************************************************/

CBulge operator * (const CBulge &bulge, const CMatrix &M)
{
	CJoint tAB = bulge.bow * M;

	if(!bulge.isArc) // degenerated to line
		return CBulge(tAB);

	FP2D tX = bulge.pointOnArc * M;

	if(tAB.A == tAB.B)
	{ // full circle
		FP2D   tR = (tAB.A - tX) / 2;
		double a  = tR.bearing();

		return CBulge(tAB.A - tR, tR.length(), a, a + 2 * M_PI);
	}

	return CBulge(tAB, tX);
}

/********************************************************************/
/* operator %: multiplies a VECTOR by a matrix.                     */
/********************************************************************/

FP2D operator % (const FP2D &V, const CMatrix &M)
{
	return V.x * M.vx + V.y * M.vy;
}

/********************************************************************/
/* PrintOn: prints the matrix on a stream.                          */
/********************************************************************/

void CMatrix::PrintOn(FILE *stream) const
{
	fprintf(stream, "( %10g %10g %10g )\n", vx.x, vy.x, va.x);
	fprintf(stream, "( %10g %10g %10g )\n", vx.y, vy.y, va.y);
}

/********************************************************************/
/* Det: computes matrix determinant.                                */
/********************************************************************/

double CMatrix::Det() const
{
	return vx % vy;
}

/********************************************************************/
/* Dist: transforms distance through the matrix.                    */
/********************************************************************/

double CMatrix::Dist(double l) const
{
	double d = Det();

	return l * (d >= 0 ? sqrt(d) : -sqrt(-d));
}

/********************************************************************/
/* Twist: transforms angle through a matrix.                        */
/********************************************************************/

double CMatrix::Twist(double a) const
{
	return (FP2D::Polar(1, a) % *this).bearing();
}
