/********************************************************************/
/* GEOMETRY: analytic geometry computations.                        */
/********************************************************************/

#ifndef __GEOMETRY_INCLUDED__
#define __GEOMETRY_INCLUDED__

#include <fp.h>
#include <stdio.h>

class CBox;
class CJoint;
class CBulge;
class CMatrix;

class CBox
{
public:
	FP2D LB, RT; /* bottomleft / topright corner */

	CBox();
	CBox(const FP2D &P, const FP2D &Q);

	void Set(const FP2D &P, const FP2D &Q);
	void operator += (const FP2D &P);

	double Width() const { return RT.x - LB.x; }
	double Height() const { return RT.x - LB.x; }

	BOOL Contains(const FP2D &X) const;
	BOOL IsNull() const;

	static CBox NIL;
};

class CJoint
{
public:
	FP2D A, B;

	CJoint();
	CJoint(const FP2D &_A, const FP2D &_B) { A = _A; B = _B; }
	CJoint(double x1, double y1, double x2, double y2)
	{
		A.x = x1; A.y = y1; B.x = x2; B.y = y2;
	}

	FP2D    Vector() const;
	void    Vector(const FP2D &V);
	double  AtY(double y) const;
	FP2D    Centre() const;
	BOOL    In(const CBox &box) const;
	FP2D    Intersect(const CJoint &L) const;
	double  Length() const;
	CJoint  Clip(const CBox &box) const;
	void    PrintOn(FILE *stream) const;
	BOOL    IsNull() const;

	static CJoint NIL;
};

class CBulge
{
public:
	CJoint bow;
	FP2D   pointOnArc;
	FP2D   centre;
	double height;
	double alpha;
	double beta;
	double radius;
	BOOL   isArc;

	CBulge() {}
	CBulge(const CJoint &bow, const FP2D &pointOnArc) { Init(bow, pointOnArc); }
	CBulge(const FP2D &X, const FP2D &Y, double height) { Init(X, Y, height); }
	CBulge(const FP2D &C, double r, double a, double b) { Init(C, r, a, b); }
	CBulge(const CJoint &bow) { Init(bow); }

	void Init(const FP2D &X, const FP2D &Y, double height);
	void Init(const CJoint &bow, const FP2D &pointOnArc);
	void Init(const FP2D &C, double r, double alpha, double beta);
	void Init(const CJoint &bow);
	void PrintOn(FILE *stream) const;
};

class CMatrix
{
public:
	FP2D vx, vy, va;

	CMatrix();
	CMatrix(const FP2D &X, const FP2D &Y);
	CMatrix(const FP2D &X, const FP2D &Y, const FP2D &A);

	static CMatrix Mirror(const CJoint &line);
	static CMatrix Move(const FP2D &V);
	static CMatrix Rotate(const FP2D &C, double ang);
	static CMatrix Scale(const FP2D &C, double t);
	static CMatrix Scale(const FP2D &C, const FP2D &S);
	static CMatrix Pitch(double angle);

	CMatrix &Keep(const FP2D &A);
	void     PrintOn(FILE *stream) const;

	double   Det() const;
	double   Dist(double d) const;
	double   Twist(double ang) const;

	friend FP2D    operator * (const FP2D &P, const CMatrix &M);
	friend CJoint  operator * (const CJoint &J, const CMatrix &M);
	friend CMatrix operator * (const CMatrix &L, const CMatrix &M);
	friend CBulge  operator * (const CBulge &B, const CMatrix &M);
	friend FP2D    operator % (const FP2D &V, const CMatrix &M);

	static CMatrix E;
};

#endif/*__GEOMETRY_INCLUDED__*/