/********************************************************************/
/* PICTEX: converts DXF files to TeX.                               */
/********************************************************************/

#include <dxfread.h>
#include <dynamem.h>
#include <bgiview.h>
#include <stdarg.h>
#include <math.h>
#include <trc.h>

/***************************** constants ****************************/

enum EnOutput { TERSE, NORMAL, VERBOSE };

/******************************* types ******************************/

struct CValue
{
public:
	double value;
	int    modify;
};

/****************************** tables ******************************/

static char usage[] =
"Usage: PICTEX { -option } <dxf source> [<tex output>]\n"
"\n"
"Options: -w[+-!]#        ..  output image width\n"
"         -h[+-!]#        ..  output image height\n"
"         -m[lrtbvh]#[%]  ..  margin (all/left/right/top/bottom/vert/horz)\n"
"         -s[xy]#         ..  scale image in both / one direction\n"
"         -r              ..  raw text (write strings into TEX file)\n"
"         -v              ..  verbose output\n"
"         -t              ..  terse output\n"
"         -axxx           ..  set arc precision to xxx mm (0.5 by default)\n"
"         -pxxx           ..  add font path xxx to path list\n"
"         -g              ..  graphic output on BGI screen\n"
;

/******************************* data *******************************/

static char     ni[80];
static char     no[80];
static EnOutput display = NORMAL;
static FP2D     picSize(100, 70); // default image: 7 x 10 cm at most
static int      sFX = -1, sFY = -1;
static FP2D     margLB(5, 5); // default margins = 5 mm
static FP2D     margRT(5, 5);
static FP2D     ratio;
static FP2D     trueSize;

/******************************* code *******************************/

/********************************************************************/
/* Normal: writes a formatted string in normal mode.                */
/********************************************************************/

static void Normal(const char *fmt, ...)
{
	if(display >= NORMAL)
	{
		va_list argp;

		va_start(argp, fmt);
		vprintf(fmt, argp);
		va_end(argp);
	}
}

/********************************************************************/
/* Verbose: displays a formatted string in verbose mode.            */
/********************************************************************/

static void Verbose(const char *fmt, ...)
{
	if(display >= VERBOSE)
	{
		va_list argp;

		va_start(argp, fmt);
		vprintf(fmt, argp);
		va_end(argp);
	}
}

/********************************************************************/
/* Progress: shows file progress.                                   */
/********************************************************************/

static void Progress(int percentage)
{
	static char *tag = "";
	char *text = format("(%s %3d%%) ", tag, percentage);
	int   back = strlen(text);

	switch(percentage)
	{
	case  -1: tag = "Sizing";     break;
	case  -2: tag = "Converting"; break;
	case   0: back = 0;
	default:
		if(display >= NORMAL)
			while(--back >= 0)
				putchar('\b');
		Normal(text);
		break;
	}
}

/********************************************************************/
/* fmax0: computes maximum from x and zero.                         */
/********************************************************************/

static double fmax0(double x)
{
	return x >= 0 ? x : 0;
}

/********************************************************************/
/* Transform: computes true picture scaling ratios.                 */
/********************************************************************/

static CMatrix Transform(const CBox &limits)
{
	if(limits.IsNull())
		return CMatrix::E;

	FP2D    size   = limits.RT - limits.LB;
	FP2D    x, y;
	FP2D    inside;
	FP2D    trueLBM;
	FP2D    trueRTM;
	FP2D    mx(
		fmax0(+margLB.x) + fmax0(+margRT.x),
		fmax0(-margLB.x) + fmax0(-margRT.x)
	);
	FP2D    my(
		fmax0(+margLB.y) + fmax0(+margRT.y),
		fmax0(-margLB.y) + fmax0(-margRT.y)
	);

	// compute picture size without margins
	if(picSize.x > 0)
	{ // compute true inside size
		if((inside.x = picSize.x * (1 - mx.y) - mx.x) <= 0)
		{
			inside.x = picSize.x;
			margLB.x = margRT.x = 0;
		}
		ratio.x = inside.x / size.x;
		if(ratio.y == 0)
			ratio.y = ratio.x;
	}

	if(picSize.y > 0)
	{ // compute true inside size
		if((inside.y = picSize.y * (1 - my.y) - my.x) <= 0)
		{
			inside.y = picSize.y;
			margLB.y = margRT.y = 0;
		}
		ratio.y = inside.y / size.y;
		if(ratio.x == 0)
			ratio.x = ratio.y;
	}

	// fix up ratios
	if(ratio.x <= 0)
		ratio.x = 1;
	if(ratio.y <= 0)
		ratio.y = 1;

	// compute size according to ratio
	if(inside.x <= 0)
	{
		sFX = 0;
		inside.x = size.x * ratio.x;
	}

	if(inside.y <= 0)
	{
		sFY = 0;
		inside.y = size.y * ratio.y;
	}

	// match sizes
	if(sFY)
	{ // match Y to X
		double ny = ratio.x * size.y;
		if(ny > inside.y && sFY > 0 || ny < inside.y && sFY < 0)
		{
			inside.y = ny;
			ratio.y  = ratio.x;
		}
	}

	if(sFX)
	{ // match X to Y
		double nx = ratio.y * size.x;
		if(nx > inside.x && sFX > 0 || nx < inside.x && sFX < 0)
		{
			inside.x = nx;
			ratio.x  = ratio.y;
		}
	}

	// compute true output size
	trueSize.x = (inside.x + mx.x) / (1 - mx.y);
	trueSize.y = (inside.y + my.x) / (1 - my.y);

	// compute margin sizes
	trueLBM = margLB;
	trueRTM = margRT;

	if(trueLBM.x < 0)  trueLBM.x *= -trueSize.x;
	if(trueLBM.y < 0)  trueLBM.y *= -trueSize.y;
	if(trueRTM.x < 0)  trueRTM.x *= -trueSize.x;
	if(trueRTM.y < 0)  trueRTM.y *= -trueSize.y;

	Verbose("Original image size: (%.3g,%.3g)\n", size.x, size.y);
	Verbose("Output image size: (%.3g,%.3g)\n", trueSize.x, trueSize.y);
	if(fabs(ratio.x - ratio.y) <= 1e-6)
		Verbose("Scaling factor = %.3g\n", ratio.x);
	else
		Verbose("Scaling factors: (%.3g,%.3g)\n", ratio.x, ratio.y);
	Verbose("Margins: L(%.3g) R(%.3g) B(%.3g) T(%.3g)\n",
		trueLBM.x, trueRTM.x, trueLBM.y, trueRTM.y);

	return CMatrix::Move(FP2D(trueLBM - limits.LB))
	* CMatrix::Scale(trueLBM, ratio);
}

/******************************* main *******************************/

int cdecl main(int, char **argv)
{
	char        *arg;
	BYTE         mask;
	double       value;
	CTexView     output;
	CLimiterView limiter;
	CDxfReader   input;
	BOOL         graph = FALSE;
	BOOL         raw_text = FALSE;
	char        *outbuf;
	double       arc_prec = 0.5;
	enum { OUTBUFSIZE = 30000 };

	{ // check ASKETCH variable & set support directory
		LPCSTR asketch = envstr("ASKETCH");
		if(asketch)
			input.AddPath(asketch);
	}

	while((arg = *++argv) != NULL)
		if(*arg == '-' || *arg == '/')
			switch(arg++, *arg++ | 0x20)
			{
			case 'a': // arc precision
				arc_prec = atof(arg);
				break;

			case 'w': // width
				sFX = -1;
				switch(*arg++)
				{
				case '+': sFX = +1; break;
				case '-': break;
				case '!': sFX = 0; break;
				default:  arg--;
				}
				picSize.x = atof(arg);
				break;

			case 'h': // height
				sFY = -1;
				switch(*arg++)
				{
				case '+': sFY = +1; break;
				case '-': break;
				case '!': sFY = 0; break;
				default:  arg--;
				}
				picSize.y = atof(arg);
				break;

			case 'm': // margin
				mask = 0xF;
				switch(*arg++ | 0x20)
				{
				case 'l': mask = 0x1; break;
				case 'r': mask = 0x2; break;
				case 'h': mask = 0x3; break;
				case 'b': mask = 0x4; break;
				case 't': mask = 0x8; break;
				case 'v': mask = 0xC; break;
				default:  arg--;
				}
				errno = 0;
				value = strtod(arg, &arg);
				if(*arg == '%')
				{
					value /= -100;
					arg++;
				}
				if(errno || *arg)
					error("invalid option syntax: '%s'", *argv);
				if(mask & 1) margLB.x = value;
				if(mask & 2) margRT.x = value;
				if(mask & 4) margLB.y = value;
				if(mask & 8) margRT.y = value;
				break;

			case 's': // scale
				mask = 3;
				switch(*arg++ | 0x20)
				{
				case 'x': mask = 1; break;
				case 'y': mask = 2; break;
				default:  arg--;
				}
				errno = 0;
				value = strtod(arg, &arg);
				if(errno || *arg)
					error("invalid option syntax: '%s'", *argv);
				if(mask & 1) ratio.x = value;
				if(mask & 2) ratio.y = value;
				break;

			case 'r': raw_text = TRUE; break;

			case 'v': display = VERBOSE; break;
			case 't': display = TERSE;   break;

			case 'p': input.AddPath(arg); break;

			case 'g': graph = TRUE; break;

			default: error("invalid option '%s'", arg - 2);
			}
		else if(*ni == 0)
			fappend(ni, arg, ".DXF");
		else if(*no == 0)
			fappend(no, arg, ".TEX");
		else
			error("no use for file '%s'", arg);

	if(*ni == 0)
	{
		printf(usage);
		return -1;
	}

	if(*no == 0)
		fforce(no, ni, ".TEX");

	if(!input.Open(ni))
		error("file '%s' - %s", ni, input.ErrName());

	Progress(-1);
	input.OnView(&limiter);
	if(!input.Expand(Progress))
		error("input file/memory failure ('%s')", ni);
	input.transform = Transform(limiter.limits);
	input.OnView(&output);
	output.penWidth = arc_prec;

	if(!output.Create(no))
		error("cannot create file '%s'", no);

	outbuf = new char[OUTBUFSIZE];
	setvbuf(output.fp, outbuf, _IOFBF, OUTBUFSIZE);
	Progress(-2);
	output.Header(trueSize);
	input.RawText(raw_text);
	if(!input.Expand(Progress))
		error("cannot process file '%s'", ni);
	output.Footer();
	if(!output.Close())
		error("cannot write file '%s'", no);
	delete[] outbuf;

	if(graph)
	{ // show graphic output
		CBgiView bgi;
		picSize = bgi.Size();
		sFX = sFY = -1;
		margLB = margRT = FP2D(20, 20);
		input.transform = Transform(limiter.limits);
		input.OnView(&bgi);
		input.RawText(FALSE);
		input.Expand();
		key();
	}
	else
		Normal("\n");

	return 0;
}
