/********************************************************************/
/* SHXFONT: interface to AutoCAD SHX font files.                    */
/********************************************************************/

#include <shxfont.h>
#include <dynamem.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

/****************************** CShx:: ******************************/

/**************************** constructor ***************************/

CShx::CShx(const char *name)
{
	file = NULL;
	font = NULL;

	FILE   *fp;
	long    fs;
	int     num;

	if((fp = fopen(name, "rb")) == NULL)
		return;
	fseek(fp, 0, SEEK_END);
	if((fs = ftell(fp)) < 24 || fs != (size_t)fs)
	ERROR:
	{ // invalid font
		fclose(fp);
		return;
	}
	file = new char[(size_t)fs];
	rewind(fp);
	if(fread(file, (size_t)fs, 1, fp) <= 0)
		goto ERROR; // file read error
	fclose(fp);
	if(memcmp(file, "AutoCAD-86 shapes 1.0\r\n\x1A", 24))
		return; // invalid font
	first = Peek(0x18);
	count = first + Peek(0x1A) + 1;
	num   = Peek(0x1C);
	font  = new char *[count];
	memset(font, 0, sizeof(char *) * count);
	char *addr = file + (0x1E + 4 * num);
	for(int i = 0, pos = 0x1E; i < num; i++, pos += 4)
	{
		int code = Peek(pos + 0);
		int size = Peek(pos + 2);

		if((unsigned)(code - first) < count)
			font[code - first] = addr + (strlen(addr) + 1);
		addr += size;
	}
	factor = 1;
	char *c0 = Chr(0);
	if(c0 != NULL) // find height of A
		if((factor = (signed char)*c0) == 0)
			factor = 1;
}

/**************************** destructor ****************************/

CShx::~CShx()
{
	if(file)
	{
		delete[] file;
		file = NULL;
	}
	if(font)
	{
		delete[] font;
		font = NULL;
	}
	count = 0;
}

/****************************** methods *****************************/

/********************************************************************/
/* IsValid: checks whether font is correctly allocated.             */
/********************************************************************/

BOOL CShx::IsValid()
{
	return file != NULL && font != NULL && count > 0;
}

/********************************************************************/
/* Peek: reads a word from the font file.                           */
/********************************************************************/

int CShx::Peek(int pos)
{
	return (unsigned char)file[pos] | ((unsigned char)file[pos + 1] << 8);
}

/********************************************************************/
/* Chr: addresses given character.                                  */
/********************************************************************/

char *CShx::Chr(int code)
{
	if(!IsValid() || (unsigned)(code - first) >= count)
		return NULL;
	return font[code - first];
}

/**************************** CShxDraw:: ****************************/

/**************************** constructor ***************************/

CShxDraw::CShxDraw(const char *name) : CShx(name)
{
	view     = NULL;
	penDown  = TRUE;
	vertical = FALSE;
	over     = FALSE;
	under    = FALSE;
}

/**************************** destructor ****************************/

CShxDraw::~CShxDraw()
{
}

/****************************** methods *****************************/

/********************************************************************/
/* MoveRel: performs a relative cursor movement or draws a line.    */
/********************************************************************/

void CShxDraw::MoveRel(const FP2D &P)
{
	if(penDown)
		view -> LineRel(P);
	else
		view -> MoveRel(P);
}

/********************************************************************/
/* Expand: expands a letter.                                        */
/********************************************************************/

void CShxDraw::Expand(int letter)
{
	static char vertShift[] = "\0\1\1\2\2\1\1\2\3\5\3\6\4\6\0\1";
	static signed char vector[][2] =
	{
		{ +2,  0 }, { +2, +1 }, { +2, +2 }, { +1, +2 },
		{  0, +2 }, { -1, +2 }, { -2, +2 }, { -2, +1 },
		{ -2,  0 }, { -2, -1 }, { -2, -2 }, { -1, -2 },
		{  0, -2 }, { +1, -2 }, { +2, -2 }, { +2, -1 },
	};
	enum { MAX_STACK = 20 }; // maximum shape nesting level

	FP2D          locStack[4];
	char         *stack[MAX_STACK];
	char         *ptr = Chr(letter);
	unsigned char c1;
	int           stp = 0; // stack pointer
	int           lsp = 0; // location stack pointer
	FP2D          bov(view -> cursor + FP2D(-0.15, +1.20));
	FP2D          bun(view -> cursor + FP2D(-0.15, -0.20));
	double        div = factor;

	if(ptr == NULL || letter == 0)
		return;
	penDown = TRUE;
	for(;;)
		if((c1 = *ptr++) >= 16)
		{ // relative vector
			int    dir = c1 & 15;
			double len = (c1 >> 4) / (2.0 * div);

			MoveRel(len * FP2D(vector[dir][0], vector[dir][1]));
		}
		else
			switch(c1)
			{
			case 0x0: // end of shape/subshape
				if(stp == 0)
				{ // end of shape
					if(over)
						view -> Line(bov, view -> cursor + FP2D(-0.15, +1.20));
					if(under)
						view -> Line(bun, view -> cursor + FP2D(-0.15, -0.20));
					return;
				}
				ptr = stack[--stp];
				break;

			case 0x1: Pen(TRUE); break;
			case 0x2: Pen(FALSE); break;
			case 0x3:
				if(*ptr)
					div *= (signed char)*ptr;
				ptr++;
				break;

			case 0x4:
				if(*ptr)
					div /= (signed char)*ptr;
				ptr++;
				break;

			case 0x5:
				if(lsp >= 4)
					return; // location stack overflow
				locStack[lsp++] = view -> cursor;
				break;

			case 0x6:
				if(lsp <= 0)
					return; // location stack underflow
				view -> MoveTo(locStack[--lsp]);
				break;

			case 0x7:
				if(stp >= MAX_STACK)
					return; // shape stack overflow
				c1 = *ptr++;
				stack[stp++] = ptr;
				if((ptr = Chr(c1)) == NULL)
					ptr = stack[--stp];
				break;

			case 0x8:
				MoveRel(FP2D((signed char)ptr[0], (signed char)ptr[1]) / div);
				ptr += 2;
				break;

			case 0x9:
				for(; ptr[0] | ptr[1]; ptr += 2)
					MoveRel(FP2D((signed char)ptr[0], (signed char)ptr[1]) / div);
				ptr += 2;
				break;

			case 0xA:
			case 0xB:
				{ // octant arc
					int    sto = 0, eno = 0, rad = 0;

					if(c1 == 0xB)
					{
						sto = (unsigned char)*ptr++;
						eno = (unsigned char)*ptr++;
						rad = 256 * (unsigned char)*ptr++;
					}
					rad |= (unsigned char)*ptr++;
					int    dir   = (signed char)*ptr++;
					int    dab   = (dir >= 0 ? dir : -dir);
					int    soc   = dab >> 4;
					int    noc   = dab & 15;

					if(noc == 0)
						noc = 8;
					if(eno)
						noc--;
					if(dir >= 0)
					{
						sto = 256 * soc + sto;
						eno = 256 * (soc + noc) + eno;
					}
					else
					{
						sto = 256 * soc - sto;
						eno = 256 * (soc - noc) - eno;
					}
					double start = M_PI / 4.0 / 256.0 * sto;
					double end   = M_PI / 4.0 / 256.0 * eno;
					CBulge arc(view -> cursor, rad / div, start, end);
					if(penDown)
						view -> Arc(arc);
					view -> MoveTo(arc.bow.B); // go to endpoint
				}
				break;

			case 0xC:
			case 0xD:
				do
				{
					FP2D v = FP2D((signed char)ptr[0], (signed char)ptr[1]) / div;
					if(c1 == 0xD && (ptr[0] | ptr[1]) == 0)
					{
						ptr += 2;
						break;
					}
					CBulge arc = CBulge(view -> cursor, view -> cursor + v,
					(signed char)ptr[2] * v.length() / 254.0);
					if(penDown)
						view -> Arc(arc);
					view -> MoveTo(arc.bow.B); // go to endpoint
					ptr += 3;
				}
				while(c1 == 0xD);
				break;

			case 0xE:
				if((c1 = *ptr) >= 0x10)
					c1 = 0x0F;
				if(!vertical)
					ptr += vertShift[c1];
			}
}

/********************************************************************/
/* Expand: expands a string. Checks %% combinations.                */
/********************************************************************/

void CShxDraw::Expand(const char *s)
{
	over  = FALSE;
	under = FALSE;
	for(unsigned char c1; (c1 = *s++) != 0;)
	{
		if(c1 == '%' && *s == '%') // %% sequence
			switch(s++, tolower(c1 = *s++))
			{
			case 0: return; // end of line
			case 'o': over  ^= 1; continue;
			case 'u': under ^= 1; continue;
			case 'd': c1 = 127; break;
			case 'p': c1 = 128; break;
			case 'c': c1 = 129; break;
			case '%': break;
			default:
				if(isdigit(*s))
				{
					c1 -= '0';
					if(isdigit(*s))
					{
						c1 = c1 * 10 + *s++ - '0';
						if(isdigit(*s))
							c1 = c1 * 10 + *s++ - '0';
					}
				}
				break;
			}
		Expand(c1);
	}
}
