/********************************************************************/
/* VIEWPORT: a generic abstract output device.                      */
/********************************************************************/

#include <viewport.h>
#include <math.h>
#include <ctype.h>
#include <stdarg.h>

/**************************** CViewport:: ***************************/

/**************************** constructor ***************************/

CViewport::CViewport()
{
	cursor.x  = 0;
	cursor.y  = 0;
	arcPrec   = DEFAULT_ARC_PREC;
	curvePrec = DEFAULT_CURVE_PREC;
	penWidth  = 1.0;
}

/**************************** destructor ****************************/

CViewport::~CViewport() {}

/****************************** methods *****************************/

/********************************************************************/
/* LineTo: draws a line segment and changes cursor position.        */
/********************************************************************/

void CViewport::LineTo(const FP2D &Q)
{
	if(cursor != Q)
	{
		Line(cursor, Q);
		cursor = Q;
	}
}

/********************************************************************/
/* Point: draws a point by painting line with zero length.          */
/********************************************************************/

void CViewport::Point(const FP2D &P)
{
	Line(P, P);
}

/********************************************************************/
/* Arc: draws an arc using line segments.                           */
/********************************************************************/

void CViewport::Arc(const CBulge &bulge)
{
	MoveTo(bulge.bow.A);
	if(!bulge.isArc)
		LineTo(bulge.bow.B);
	else
	{
		if(bulge.bow.A != bulge.bow.B)
			RecurseArc(bulge.bow.A, bulge.bow.B,
				bulge.bow.Length(), bulge.height,
				arcPrec);
		else
		{ // full circle - draw using 2 recursed arcs
			FP2D B(2 * bulge.centre - bulge.bow.A);
			RecurseArc(bulge.bow.A, B, 2 * bulge.radius, bulge.radius, arcPrec);
			RecurseArc(B, bulge.bow.A, 2 * bulge.radius, bulge.radius, arcPrec);
		}
	}
}

/********************************************************************/
/* RecurseArc: draws arc with line segments using recursion.        */
/********************************************************************/

void CViewport::RecurseArc(const FP2D &A, const FP2D &B, double l, double h, int n)
{
	if(h <= penWidth || --n < 0)
		LineTo(B);
	else
	{
		FP2D   M  = (A + B) / 2 + (B - A).right() * (h / l);
		double l2 = sqrt(l * l / 4 + h * h);
		double h2 = (l * l + 4 * h * h - 2 * l * l2) / (8 * h);

		RecurseArc(A, M, l2, h2, n);
		RecurseArc(M, B, l2, h2, n);
	}
}

/********************************************************************/
/* Circle: draws a circle.                                          */
/********************************************************************/

void CViewport::Circle(const FP2D &C, double r)
{
	Arc(CBulge(C, r, 0, 2 * M_PI));
}

/********************************************************************/
/* CurveTo: draws 2nd order Bezier curve from cursor.               */
/********************************************************************/

void CViewport::CurveTo(const FP2D &E, const FP2D &M)
{
	FP2D   kt   = 2 * (M - E);
	FP2D   kt2  = cursor - E - kt;

	for(int i = 1; i < curvePrec; i++)
	{
		double t = i / (double)curvePrec;
		LineTo(E + kt * t + kt2 * (t * t));
	}
	LineTo(E);
}

/********************************************************************/
/* Text: displays a text with given properties.                     */
/********************************************************************/

#pragma argsused
void CViewport::Text(const FP2D &insert, double length, double height,
double angle, double pitch, int justif, const char *text)
{ // currently not implemented for base class
}

/**************************** CNilView:: ****************************/

/**************************** static data ***************************/

CNilView *CNilView::O;

/************************ CTransformedView:: ************************/

/**************************** constructor ***************************/

CTransformedView::CTransformedView(CViewport *view, const CMatrix &matrix)
: view(view), matrix(matrix)
{
	UpdatePenWidth();
}

/****************************** methods *****************************/

void CTransformedView::UpdatePenWidth()
{
	if(view != NULL)
		penWidth = view -> penWidth / matrix.Dist(1);
}

/********************************************************************/
/* Point: sends scaled point to slave view.                         */
/********************************************************************/

void CTransformedView::Point(const FP2D &P)
{
	view -> Point(P * matrix);
}

/********************************************************************/
/* Line: sends scaled line to slave view.                           */
/********************************************************************/

void CTransformedView::Line(const FP2D &P, const FP2D &Q)
{
	view -> Line(P * matrix, Q * matrix);
}

/********************************************************************/
/* MoveTo: updates both original and scaled cursor.                 */
/********************************************************************/

void CTransformedView::MoveTo(const FP2D &P)
{
	view -> MoveTo(P * matrix);
	cursor = P;
}

/********************************************************************/
/* LineTo: draws a line segment and updates both cursors.           */
/********************************************************************/

void CTransformedView::LineTo(const FP2D &P)
{
	view -> LineTo(P * matrix);
	cursor = P;
}

/********************************************************************/
/* Text: draws transformed text.                                    */
/********************************************************************/

void CTransformedView::Text(const FP2D &insert, double width, double height,
double angle, double pitch, int justif, const char *text)
{
	view -> Text(insert * matrix, matrix.Dist(width), matrix.Dist(height),
	matrix.Twist(angle), pitch, justif, text);
}

/************************** CLimiterView:: **************************/

/**************************** constructor ***************************/

CLimiterView::CLimiterView()
{
	limits = CBox::NIL;
}

/****************************** methods *****************************/

/********************************************************************/
/* Line: includes the line into limits.                             */
/********************************************************************/

void CLimiterView::Line(const FP2D &P, const FP2D &Q)
{
	limits += P;
	limits += Q;
}

/**************************** CTexView:: ****************************/

/**************************** constructor ***************************/

CTexView::CTexView()
{
	fp     = NULL;
	prec   = 100;
}

/**************************** destructor ****************************/

CTexView::~CTexView()
{
	Close();
}

/****************************** methods *****************************/

/********************************************************************/
/* Create: creates a TeX file.                                      */
/********************************************************************/

BOOL CTexView::Create(const char *fn)
{
	Close();
	last   = FP2D::NIL;
	column = 0;
	if((fp = fopen(fn, "wb")) == NULL)
		return FALSE;
	return TRUE;
}

/********************************************************************/
/* Close: closes the TeX file.                                      */
/********************************************************************/

BOOL CTexView::Close()
{
	int res = fclose(fp);
	fp = NULL;
	return !res;
}

/********************************************************************/
/* Point: writes a point to the TeX file.                           */
/********************************************************************/

void CTexView::Point(const FP2D &P)
{
	Append("\\qP(%.0f,%.0f)", prec * P.x, prec * P.y);
}

/********************************************************************/
/* Line: writes a line into the TeX file.                           */
/********************************************************************/

void CTexView::Line(const FP2D &P, const FP2D &Q)
{
	FP2D tP = FP2D(floor(0.5 + P.x * prec), floor(0.5 + P.y * prec));
	FP2D tQ = FP2D(floor(0.5 + Q.x * prec), floor(0.5 + Q.y * prec));
	if(tP != tQ)
	{
		if(tQ == last)
			memswap(&tP, &tQ, sizeof(FP2D));
		if(tP != last)
			Append("\\qM(%.0f,%.0f)", tP.x, tP.y);
		Append("\\qL(%.0f,%.0f)", tQ.x, tQ.y);
		last = tQ;
	}
	else if(P == Q) // dot
		Append("\\qP(%.0f,%.0f)", tP.x, tP.y);
}

/********************************************************************/
/* Text: writes a text block into the TeX file.                     */
/********************************************************************/

void CTexView::Text
(
	const FP2D &insert, double width, double height,
	double /*angle*/, double pitch, int justif, const char *text
)
{
	// compute bounding box
	FP2D bottom_left = insert;

	switch(justif & 007)
	{
	default:
	case 000: break;
	case 001: bottom_left.x -= width / 2; break;
	case 002: bottom_left.x -= width; break;
	}

	switch(justif & 070)
	{
	default:
	case 000: break;
	case 010: bottom_left.y -= height / 2; break;
	case 020: bottom_left.y -= height; break;
	}

	if(column)
		fputs("\r\n", fp);
	fprintf(fp, "\\put(%.0f,%.0f){\\makebox(%.0f,%.0f)[%c%c]{",
	prec * bottom_left.x, prec * bottom_left.y, prec * width, prec * height,
	"lcr"[(justif >> 0) & 7], "bct"[(justif >> 3) & 7]);
	if(pitch > 0)
		fputs("\\it ", fp);

	{ // write font size
		const char *size = Fontsize(height);
		if(size)
		{
			fputc('\\', fp);
			fputs(size, fp);
			fputc(' ', fp);
		}
	}

	for(char c1; (c1 = *text++) != 0;)
		if(c1 >= ' ' && c1 <= '~')
		{ // currently support only ASCII characters
			if(c1 == '%' && *text == '%')
			{
				text++;
				switch(tolower(c1 = *text++))
				{
				case 0: // end of text
					goto END_TEXT;

				case 'o': // overline
				case 'u': // underline
					continue;

				case 'd': // degree
					fputs("${}^0$", fp);
					continue;

				case 'p': // plusminus
					fputs("$\\pm$", fp);
					continue;

				case 'c': // diameter
					fputs("$\\oslash$", fp);
					continue;

				case '%': // '%' character
					break;

				default:
					if(isdigit(c1))
					{
						c1 -= '0';
						if(isdigit(*text))
						{
							c1 = c1 * 10 + *text++ - '0';
							if(isdigit(*text))
								c1 = c1 * 10 + *text++ - '0';
						}
					}
					break;
				}
			}

			switch(c1)
			{
			case '%': case '#': case '$': case '&': case '{': case '}':
				fputc('\\', fp);
				fputc(c1, fp);
				break;
			case '<': case '>': case '|':
				fprintf(fp, "$%c$", c1);
				break;
			case '_': case '^': case '~':
				fprintf(fp, "\\%c{}", c1);
				break;
			case '\\':
				fputs("$\\backslash$", fp);
				break;
			default:
				fputc(c1, fp);
				break;
			}
		}

END_TEXT:
	fputs("}}\r\n", fp);
	column = 0;
}

/********************************************************************/
/* Header: writes TeX file heading.                                 */
/********************************************************************/

void CTexView::Header(const FP2D &S)
{
	// ? write linewidth item ?
	fprintf(fp, "\\def\\qP(#1,#2){\\put(#1,#2){\\circle{0}}}\r\n");
	fprintf(fp, "\\def\\qM(#1,#2){\\put(#1,#2){\\special{em:moveto}}}\r\n");
	fprintf(fp, "\\def\\qL(#1,#2){\\put(#1,#2){\\special{em:lineto}}}\r\n");
	fprintf(fp, "\\setlength{\\unitlength}{%gmm}\r\n", 1.0 / prec);
	fprintf(fp, "\\begin{picture}(%.0f,%.0f)\r\n", prec * S.x, prec * S.y);
}

/********************************************************************/
/* Footer: closes TeX sequence.                                     */
/********************************************************************/

void CTexView::Footer()
{
	fprintf(fp, "\r\n\\end{picture}\r\n");
}

/********************************************************************/
/* Append: writes an arbitrary formatted text. Performs wordwrap.   */
/********************************************************************/

void CTexView::Append(const char *fmt, ...)
{
	static char buffer[1000];

	va_list args;
	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);
	int length = strlen(buffer);
	if(length + column > 75)
	{
		fputs("\r\n", fp);
		column = 0;
	}
	fputs(buffer, fp);
	column += length;
}

/********************************************************************/
/* Fontsize: converts font height to letter size.                   */
/********************************************************************/

const char *CTexView::Fontsize(double height)
{
	struct TSize
	{
		int         limit; // max. size in 1/1000 mm
		const char *name;  // font name
	};

	static TSize fontList[] =
	{
		{ 5, "tiny"         },
		{ 7, "scriptsize"   },
		{ 8, "footnotesize" },
		{ 9, "small"        },
		{ 10, NULL           }, // normal
		{ 12, "large"        },
		{ 14, "Large"        },
		{ 17, "LARGE"        },
		{ 20, "huge"         },
		{ 25, "Huge"         },
		{ 0 }
	};

	height *= 5; // approximate scaling factor to points
	for(TSize *ptr = fontList; ptr[1].limit && ptr[1].limit < height; ptr++)
		;
	return ptr -> name;
}
