/********************************************************************/
/* VIEWPORT: a generic abstract output device.                      */
/********************************************************************/

#ifndef __VIEWPORT_INCLUDED__
#define __VIEWPORT_INCLUDED__

#include <geometry.h>

class CViewport
{
public:
	FP2D   cursor;
	int    arcPrec;
	int    curvePrec;
	double penWidth;
	enum
	{
		DEFAULT_ARC_PREC   = 6,
		DEFAULT_CURVE_PREC = 6,
	};

	CViewport();
	virtual ~CViewport();

	virtual void Line(const FP2D &P, const FP2D &Q) = 0;
	virtual void MoveTo(const FP2D &P) { cursor = P; }
	virtual void MoveRel(const FP2D &P) { MoveTo(cursor + P); }
	virtual void LineTo(const FP2D &P);
	virtual void LineRel(const FP2D &P) { LineTo(cursor + P); }
	virtual void Point(const FP2D &P);
	virtual void Arc(const CBulge &bulge);
	virtual void Circle(const FP2D &C, double r);
	virtual void CurveTo(const FP2D &E, const FP2D &M);
	virtual void Text(const FP2D &insert, double length, double height,
	double angle, double pitch, int justif, const char *text);

protected:
	void RecurseArc(const FP2D &A, const FP2D &B, double l, double h, int n);
};

class CNilView : public CViewport
{
public:
	virtual void Line(const FP2D &P, const FP2D &Q) {}

	static CNilView *O;
};

class CTransformedView : public CViewport
{
protected:
	CMatrix    matrix;
	CViewport *view;

public:
	CTransformedView(CViewport *_view, const CMatrix &_matrix);

	void UpdatePenWidth();
	void View(CViewport *_view) { view = _view; UpdatePenWidth(); }
	CViewport *View() { return view; }

	void Matrix(const CMatrix &_matrix) { matrix = _matrix; UpdatePenWidth(); }
	const CMatrix &Matrix() { return matrix; }

	virtual void Line(const FP2D &P, const FP2D &Q);
	virtual void MoveTo(const FP2D &P);
	virtual void LineTo(const FP2D &P);
	virtual void Point(const FP2D &P);
	virtual void Text(const FP2D &insert, double length, double height,
	double angle, double pitch, int justif, const char *text);
};

class CLimiterView : public CViewport
{
public:
	CBox limits;

	CLimiterView();

	virtual void Line(const FP2D &P, const FP2D &Q);
};

class CTexView : public CViewport
{
private:
	void Append(const char *fmt, ...);
	static const char *Fontsize(double height);

public:
	FILE   *fp;
	double  prec;
	FP2D    last;
	int     column;

	CTexView();
	virtual ~CTexView();

	BOOL Create(const char *fn);
	BOOL Close();

	virtual void Point(const FP2D &P);
	virtual void Line(const FP2D &P, const FP2D &Q);
	virtual void Text(const FP2D &insert, double length, double height,
	double angle, double pitch, int justif, const char *text);

	void Header(const FP2D &size);
	void Footer();

	void   SetPrec(double _prec) { prec = _prec; }
	double GetPrec() const       { return prec; }
};

#endif/*__VIEWPORT_INCLUDED__*/